@@ -11,8 +11,11 @@ import binascii
 import datetime
 import errno
 import gettext
+import hashlib
+import hmac
 import importlib
 import io
+import ipaddress
 import json
 import logging
 import mimetypes
@@ -24,6 +27,7 @@ import signal
 import socket
 import sys
 import threading
+import time
 import warnings
 import webbrowser
 
@@ -102,7 +106,7 @@ from ipython_genutils import py3compat
 from jupyter_core.paths import jupyter_runtime_dir, jupyter_path
 from notebook._sysinfo import get_sys_info
 
-from ._tz import utcnow
+from ._tz import utcnow, utcfromtimestamp
 from .utils import url_path_join, check_pid, url_escape
 
 #-----------------------------------------------------------------------------
@@ -132,7 +136,6 @@ def random_ports(port, n):
 
 def load_handlers(name):
     """Load the (URL pattern, handler) tuples for each component."""
-    name = 'notebook.' + name
     mod = __import__(name, fromlist=['default_handlers'])
     return mod.default_handlers
 
@@ -144,21 +147,22 @@ class NotebookWebApplication(web.Application):
 
     def __init__(self, jupyter_app, kernel_manager, contents_manager,
                  session_manager, kernel_spec_manager,
-                 config_manager, log,
+                 config_manager, extra_services, log,
                  base_url, default_url, settings_overrides, jinja_env_options):
 
 
         settings = self.init_settings(
             jupyter_app, kernel_manager, contents_manager,
-            session_manager, kernel_spec_manager, config_manager, log,
-            base_url, default_url, settings_overrides, jinja_env_options)
+            session_manager, kernel_spec_manager, config_manager,
+            extra_services, log, base_url,
+            default_url, settings_overrides, jinja_env_options)
         handlers = self.init_handlers(settings)
 
         super(NotebookWebApplication, self).__init__(handlers, **settings)
 
     def init_settings(self, jupyter_app, kernel_manager, contents_manager,
                       session_manager, kernel_spec_manager,
-                      config_manager,
+                      config_manager, extra_services,
                       log, base_url, default_url, settings_overrides,
                       jinja_env_options=None):
 
@@ -230,7 +234,7 @@ class NotebookWebApplication(web.Application):
             },
             version_hash=version_hash,
             ignore_minified_js=jupyter_app.ignore_minified_js,
-            
+
             # rate limits
             iopub_msg_rate_limit=jupyter_app.iopub_msg_rate_limit,
             iopub_data_rate_limit=jupyter_app.iopub_data_rate_limit,
@@ -240,7 +244,7 @@ class NotebookWebApplication(web.Application):
             # tornado defaults are 100 MiB, we increase it to 0.5 GiB
             max_body_size = 512 * 1024 * 1024,
             max_buffer_size = 512 * 1024 * 1024,
-            
+
             # authentication
             cookie_secret=jupyter_app.cookie_secret,
             login_url=url_path_join(base_url,'/login'),
@@ -249,6 +253,8 @@ class NotebookWebApplication(web.Application):
             password=jupyter_app.password,
             xsrf_cookies=True,
             disable_check_xsrf=jupyter_app.disable_check_xsrf,
+            allow_remote_access=jupyter_app.allow_remote_access,
+            local_hostnames=jupyter_app.local_hostnames,
 
             # managers
             kernel_manager=kernel_manager,
@@ -257,15 +263,23 @@ class NotebookWebApplication(web.Application):
             kernel_spec_manager=kernel_spec_manager,
             config_manager=config_manager,
 
+            # handlers
+            extra_services=extra_services,
+
             # Jupyter stuff
             started=now,
+            # place for extensions to register activity
+            # so that they can prevent idle-shutdown
+            last_activity_times={},
             jinja_template_vars=jupyter_app.jinja_template_vars,
             nbextensions_path=jupyter_app.nbextensions_path,
             websocket_url=jupyter_app.websocket_url,
             mathjax_url=jupyter_app.mathjax_url,
             mathjax_config=jupyter_app.mathjax_config,
+            shutdown_button=jupyter_app.quit_button,
             config=jupyter_app.config,
             config_dir=jupyter_app.config_dir,
+            allow_password_change=jupyter_app.allow_password_change,
             server_root_dir=root_dir,
             jinja2_env=env,
             terminals_available=False,  # Set later if terminals are available
@@ -280,25 +294,28 @@ class NotebookWebApplication(web.Application):
 
         # Order matters. The first handler to match the URL will handle the request.
         handlers = []
-        handlers.extend(load_handlers('tree.handlers'))
+        # load extra services specified by users before default handlers
+        for service in settings['extra_services']:
+            handlers.extend(load_handlers(service))
+        handlers.extend(load_handlers('notebook.tree.handlers'))
         handlers.extend([(r"/login", settings['login_handler_class'])])
         handlers.extend([(r"/logout", settings['logout_handler_class'])])
-        handlers.extend(load_handlers('files.handlers'))
-        handlers.extend(load_handlers('view.handlers'))
-        handlers.extend(load_handlers('notebook.handlers'))
-        handlers.extend(load_handlers('nbconvert.handlers'))
-        handlers.extend(load_handlers('bundler.handlers'))
-        handlers.extend(load_handlers('kernelspecs.handlers'))
-        handlers.extend(load_handlers('edit.handlers'))
-        handlers.extend(load_handlers('services.api.handlers'))
-        handlers.extend(load_handlers('services.config.handlers'))
-        handlers.extend(load_handlers('services.kernels.handlers'))
-        handlers.extend(load_handlers('services.contents.handlers'))
-        handlers.extend(load_handlers('services.sessions.handlers'))
-        handlers.extend(load_handlers('services.nbconvert.handlers'))
-        handlers.extend(load_handlers('services.kernelspecs.handlers'))
-        handlers.extend(load_handlers('services.security.handlers'))
-        handlers.extend(load_handlers('services.shutdown'))
+        handlers.extend(load_handlers('notebook.files.handlers'))
+        handlers.extend(load_handlers('notebook.view.handlers'))
+        handlers.extend(load_handlers('notebook.notebook.handlers'))
+        handlers.extend(load_handlers('notebook.nbconvert.handlers'))
+        handlers.extend(load_handlers('notebook.bundler.handlers'))
+        handlers.extend(load_handlers('notebook.kernelspecs.handlers'))
+        handlers.extend(load_handlers('notebook.edit.handlers'))
+        handlers.extend(load_handlers('notebook.services.api.handlers'))
+        handlers.extend(load_handlers('notebook.services.config.handlers'))
+        handlers.extend(load_handlers('notebook.services.kernels.handlers'))
+        handlers.extend(load_handlers('notebook.services.contents.handlers'))
+        handlers.extend(load_handlers('notebook.services.sessions.handlers'))
+        handlers.extend(load_handlers('notebook.services.nbconvert.handlers'))
+        handlers.extend(load_handlers('notebook.services.kernelspecs.handlers'))
+        handlers.extend(load_handlers('notebook.services.security.handlers'))
+        handlers.extend(load_handlers('notebook.services.shutdown'))
         handlers.extend(settings['contents_manager'].get_extra_handlers())
 
         handlers.append(
@@ -314,7 +331,7 @@ class NotebookWebApplication(web.Application):
             })
         )
         # register base handlers last
-        handlers.extend(load_handlers('base.handlers'))
+        handlers.extend(load_handlers('notebook.base.handlers'))
         # set the URL that will be redirected from `/`
         handlers.append(
             (r'/?', RedirectWithParams, {
@@ -333,6 +350,27 @@ class NotebookWebApplication(web.Application):
         new_handlers.append((r'(.*)', Template404))
         return new_handlers
 
+    def last_activity(self):
+        """Get a UTC timestamp for when the server last did something.
+
+        Includes: API activity, kernel activity, kernel shutdown, and terminal
+        activity.
+        """
+        sources = [
+            self.settings['started'],
+            self.settings['kernel_manager'].last_kernel_activity,
+        ]
+        try:
+            sources.append(self.settings['api_last_activity'])
+        except KeyError:
+            pass
+        try:
+            sources.append(self.settings['terminal_last_activity'])
+        except KeyError:
+            pass
+        sources.extend(self.settings['last_activity_times'].values())
+        return max(sources)
+
 
 class NotebookPasswordApp(JupyterApp):
     """Set a password for the notebook server.
@@ -351,6 +389,51 @@ class NotebookPasswordApp(JupyterApp):
         set_password(config_file=self.config_file)
         self.log.info("Wrote hashed password to %s" % self.config_file)
 
+def shutdown_server(server_info, timeout=5, log=None):
+    """Shutdown a notebook server in a separate process.
+
+    *server_info* should be a dictionary as produced by list_running_servers().
+
+    Will first try to request shutdown using /api/shutdown .
+    On Unix, if the server is still running after *timeout* seconds, it will
+    send SIGTERM. After another timeout, it escalates to SIGKILL.
+
+    Returns True if the server was stopped by any means, False if stopping it
+    failed (on Windows).
+    """
+    from tornado.httpclient import HTTPClient, HTTPRequest
+    url = server_info['url']
+    pid = server_info['pid']
+    req = HTTPRequest(url + 'api/shutdown', method='POST', body=b'', headers={
+        'Authorization': 'token ' + server_info['token']
+    })
+    if log: log.debug("POST request to %sapi/shutdown", url)
+    HTTPClient().fetch(req)
+
+    # Poll to see if it shut down.
+    for _ in range(timeout*10):
+        if check_pid(pid):
+            if log: log.debug("Server PID %s is gone", pid)
+            return True
+        time.sleep(0.1)
+
+    if sys.platform.startswith('win'):
+        return False
+
+    if log: log.debug("SIGTERM to PID %s", pid)
+    os.kill(pid, signal.SIGTERM)
+
+    # Poll to see if it shut down.
+    for _ in range(timeout * 10):
+        if check_pid(pid):
+            if log: log.debug("Server PID %s is gone", pid)
+            return True
+        time.sleep(0.1)
+
+    if log: log.debug("SIGKILL to PID %s", pid)
+    os.kill(pid, signal.SIGKILL)
+    return True  # SIGKILL cannot be caught
+
 
 class NbserverStopApp(JupyterApp):
     version = __version__
@@ -364,14 +447,18 @@ class NbserverStopApp(JupyterApp):
         if self.extra_args:
             self.port=int(self.extra_args[0])
 
+    def shutdown_server(self, server):
+        return shutdown_server(server, log=self.log)
+
     def start(self):
         servers = list(list_running_servers(self.runtime_dir))
         if not servers:
             self.exit("There are no running servers")
         for server in servers:
             if server['port'] == self.port:
-                self.log.debug("Shutting down notebook server with PID: %i", server['pid'])
-                os.kill(server['pid'], signal.SIGTERM)
+                print("Shutting down server on port", self.port, "...")
+                if not self.shutdown_server(server):
+                    sys.exit("Could not stop server")
                 return
         else:
             print("There is currently no server running on port {}".format(self.port), file=sys.stderr)
@@ -582,6 +669,19 @@ class NotebookApp(JupyterApp):
             value = u''
         return value
 
+    custom_display_url = Unicode(u'', config=True,
+        help=_("""Override URL shown to users.
+
+        Replace actual URL, including protocol, address, port and base URL,
+        with the given value when displaying URL to the users. Do not change
+        the actual connection URL. If authentication token is enabled, the
+        token is added to the custom URL automatically.
+
+        This option is intended to be used when the URL to display to the user
+        cannot be determined reliably by the Jupyter notebook server (proxified
+        or containerized setups for example).""")
+    )
+
     port = Integer(8888, config=True,
         help=_("The port the notebook server will listen on.")
     )
@@ -609,32 +709,38 @@ class NotebookApp(JupyterApp):
     @default('cookie_secret_file')
     def _default_cookie_secret_file(self):
         return os.path.join(self.runtime_dir, 'notebook_cookie_secret')
-    
+
     cookie_secret = Bytes(b'', config=True,
         help="""The random bytes used to secure cookies.
         By default this is a new random number every time you start the Notebook.
         Set it to a value in a config file to enable logins to persist across server sessions.
-        
+
         Note: Cookie secrets should be kept private, do not share config files with
         cookie_secret stored in plaintext (you can read the value from a file).
         """
     )
-    
+
     @default('cookie_secret')
     def _default_cookie_secret(self):
         if os.path.exists(self.cookie_secret_file):
             with io.open(self.cookie_secret_file, 'rb') as f:
-                return f.read()
+                key =  f.read()
         else:
-            secret = encodebytes(os.urandom(1024))
-            self._write_cookie_secret_file(secret)
-            return secret
-    
+            key = encodebytes(os.urandom(32))
+            self._write_cookie_secret_file(key)
+        h = hmac.new(key, digestmod=hashlib.sha256)
+        h.update(self.password.encode())
+        return h.digest()
+
     def _write_cookie_secret_file(self, secret):
         """write my secret to my secret_file"""
         self.log.info(_("Writing notebook server cookie secret to %s"), self.cookie_secret_file)
-        with io.open(self.cookie_secret_file, 'wb') as f:
-            f.write(secret)
+        try:
+            with io.open(self.cookie_secret_file, 'wb') as f:
+                f.write(secret)
+        except OSError as e:
+            self.log.error(_("Failed to write cookie secret to %s: %s"),
+                           self.cookie_secret_file, e)
         try:
             os.chmod(self.cookie_secret_file, 0o600)
         except OSError:
@@ -700,6 +806,18 @@ class NotebookApp(JupyterApp):
                       """
     )
 
+    allow_password_change = Bool(True, config=True, 
+                    help="""Allow password to be changed at login for the notebook server. 
+
+                    While loggin in with a token, the notebook server UI will give the opportunity to
+                    the user to enter a new password at the same time that will replace
+                    the token login mechanism. 
+
+                    This can be set to false to prevent changing password from the UI/API.
+                    """
+    )
+
+
     disable_check_xsrf = Bool(False, config=True,
         help="""Disable cross-site-request-forgery protection
 
@@ -716,6 +834,50 @@ class NotebookApp(JupyterApp):
         """
     )
 
+    allow_remote_access = Bool(config=True,
+       help="""Allow requests where the Host header doesn't point to a local server
+
+       By default, requests get a 403 forbidden response if the 'Host' header
+       shows that the browser thinks it's on a non-local domain.
+       Setting this option to True disables this check.
+
+       This protects against 'DNS rebinding' attacks, where a remote web server
+       serves you a page and then changes its DNS to send later requests to a
+       local IP, bypassing same-origin checks.
+
+       Local IP addresses (such as 127.0.0.1 and ::1) are allowed as local,
+       along with hostnames configured in local_hostnames.
+       """)
+
+    @default('allow_remote_access')
+    def _default_allow_remote(self):
+        """Disallow remote access if we're listening only on loopback addresses"""
+        # Disable the check temporarily because of Mac issues:
+        # https://github.com/jupyter/notebook/issues/3754
+        return True
+
+        try:
+            addr = ipaddress.ip_address(self.ip)
+        except ValueError:
+            # Address is a hostname
+            for info in socket.getaddrinfo(self.ip, self.port, 0, socket.SOCK_STREAM):
+                addr = info[4][0]
+                if not py3compat.PY3:
+                    addr = addr.decode('ascii')
+                if not ipaddress.ip_address(addr).is_loopback:
+                    return True
+            return False
+        else:
+            return not addr.is_loopback
+
+    local_hostnames = List(Unicode(), ['localhost'], config=True,
+       help="""Hostnames to allow as local when allow_remote_access is False.
+
+       Local IP addresses (such as 127.0.0.1 and ::1) are automatically accepted
+       as local as well.
+       """
+    )
+
     open_browser = Bool(True, config=True,
                         help="""Whether to open in a browser after starting.
                         The specific browser used is platform dependent and
@@ -737,9 +899,11 @@ class NotebookApp(JupyterApp):
         `new` argument passed to the standard library method `webbrowser.open`.
         The behaviour is not guaranteed, but depends on browser support. Valid
         values are:
-            2 opens a new tab,
-            1 opens a new window,
-            0 opens in an existing window.
+
+         - 2 opens a new tab,
+         - 1 opens a new window,
+         - 0 opens in an existing window.
+
         See the `webbrowser.open` documentation for details.
         """))
 
@@ -864,6 +1028,10 @@ class NotebookApp(JupyterApp):
     extra_nbextensions_path = List(Unicode(), config=True,
         help=_("""extra paths to look for Javascript notebook extensions""")
     )
+
+    extra_services = List(Unicode(), config=True,
+        help=_("""handlers that should be loaded at higher priority than the default services""")
+    )
     
     @property
     def nbextensions_path(self):
@@ -916,6 +1084,11 @@ class NotebookApp(JupyterApp):
     @observe('mathjax_config')
     def _update_mathjax_config(self, change):
         self.log.info(_("Using MathJax configuration file: %s"), change['new'])
+        
+    quit_button = Bool(True, config=True,
+        help="""If True, display a button in the dashboard to quit
+        (shutdown the notebook server)."""
+    )
 
     contents_manager_class = Type(
         default_value=LargeFileManager,
@@ -1070,6 +1243,26 @@ class NotebookApp(JupyterApp):
     rate_limit_window = Float(3, config=True, help=_("""(sec) Time window used to 
         check the message and data rate limits."""))
 
+    shutdown_no_activity_timeout = Integer(0, config=True,
+        help=("Shut down the server after N seconds with no kernels or "
+              "terminals running and no activity. "
+              "This can be used together with culling idle kernels "
+              "(MappingKernelManager.cull_idle_timeout) to "
+              "shutdown the notebook server when it's not in use. This is not "
+              "precisely timed: it may shut down up to a minute later. "
+              "0 (the default) disables this automatic shutdown.")
+    )
+
+    terminals_enabled = Bool(True, config=True,
+         help=_("""Set to False to disable terminals.
+
+         This does *not* make the notebook server more secure by itself.
+         Anything the user can in a terminal, they can also do in a notebook.
+
+         Terminals may also be automatically disabled if the terminado package
+         is not available.
+         """))
+
     def parse_command_line(self, argv=None):
         super(NotebookApp, self).parse_command_line(argv)
 
@@ -1156,9 +1349,9 @@ class NotebookApp(JupyterApp):
         self.web_app = NotebookWebApplication(
             self, self.kernel_manager, self.contents_manager,
             self.session_manager, self.kernel_spec_manager,
-            self.config_manager,
+            self.config_manager, self.extra_services,
             self.log, self.base_url, self.default_url, self.tornado_settings,
-            self.jinja_environment_options
+            self.jinja_environment_options,
         )
         ssl_options = self.ssl_options
         if self.certfile:
@@ -1206,8 +1399,16 @@ class NotebookApp(JupyterApp):
     
     @property
     def display_url(self):
-        ip = self.ip if self.ip else _('[all ip addresses on your system]')
-        url = self._url(ip)
+        if self.custom_display_url:
+            url = self.custom_display_url
+            if not url.endswith('/'):
+                url += '/'
+        else:
+            if self.ip in ('', '0.0.0.0'):
+                ip = "(%s or 127.0.0.1)" % socket.gethostname()
+            else:
+                ip = self.ip
+            url = self._url(ip)
         if self.token:
             # Don't log full token if it came from config
             token = self.token if self._token_generated else '...'
@@ -1224,13 +1425,15 @@ class NotebookApp(JupyterApp):
         return "%s://%s:%i%s" % (proto, ip, self.port, self.base_url)
 
     def init_terminals(self):
+        if not self.terminals_enabled:
+            return
+
         try:
             from .terminal import initialize
             initialize(self.web_app, self.notebook_dir, self.connection_url, self.terminado_settings)
             self.web_app.settings['terminals_available'] = True
         except ImportError as e:
-            log = self.log.debug if sys.platform == 'win32' else self.log.warning
-            log(_("Terminals not available (error was %s)"), e)
+            self.log.warning(_("Terminals not available (error was %s)"), e)
 
     def init_signal(self):
         if not sys.platform.startswith('win') and sys.stdin and sys.stdin.isatty():
@@ -1357,6 +1560,37 @@ class NotebookApp(JupyterApp):
         # mimetype always needs to be text/css, so we override it here.
         mimetypes.add_type('text/css', '.css')
 
+
+    def shutdown_no_activity(self):
+        """Shutdown server on timeout when there are no kernels or terminals."""
+        km = self.kernel_manager
+        if len(km) != 0:
+            return   # Kernels still running
+
+        try:
+            term_mgr = self.web_app.settings['terminal_manager']
+        except KeyError:
+            pass  # Terminals not enabled
+        else:
+            if term_mgr.terminals:
+                return   # Terminals still running
+
+        seconds_since_active = \
+            (utcnow() - self.web_app.last_activity()).total_seconds()
+        self.log.debug("No activity for %d seconds.",
+                       seconds_since_active)
+        if seconds_since_active > self.shutdown_no_activity_timeout:
+            self.log.info("No kernels or terminals for %d seconds; shutting down.",
+                          seconds_since_active)
+            self.stop()
+
+    def init_shutdown_no_activity(self):
+        if self.shutdown_no_activity_timeout > 0:
+            self.log.info("Will shut down after %d seconds with no kernels or terminals.",
+                          self.shutdown_no_activity_timeout)
+            pc = ioloop.PeriodicCallback(self.shutdown_no_activity, 60000)
+            pc.start()
+
     @catch_config_error
     def initialize(self, argv=None):
         super(NotebookApp, self).initialize(argv)
@@ -1370,6 +1604,7 @@ class NotebookApp(JupyterApp):
         self.init_signal()
         self.init_server_extensions()
         self.init_mime_overrides()
+        self.init_shutdown_no_activity()
 
     def cleanup_kernels(self):
         """Shutdown all kernels.
@@ -1382,13 +1617,14 @@ class NotebookApp(JupyterApp):
         self.log.info(kernel_msg % n_kernels)
         self.kernel_manager.shutdown_all()
 
-    def notebook_info(self):
+    def notebook_info(self, kernel_count=True):
         "Return the current working directory and the server url information"
         info = self.contents_manager.info_string() + "\n"
-        n_kernels = len(self.kernel_manager.list_kernel_ids())
-        kernel_msg = trans.ngettext("%d active kernel", "%d active kernels", n_kernels)
-        info += kernel_msg % n_kernels
-        info += "\n"
+        if kernel_count:
+            n_kernels = len(self.kernel_manager.list_kernel_ids())
+            kernel_msg = trans.ngettext("%d active kernel", "%d active kernels", n_kernels)
+            info += kernel_msg % n_kernels
+            info += "\n"
         # Format the info so that the URL fits on a single line in 80 char display
         info += _("The Jupyter Notebook is running at:\n%s") % self.display_url
         return info
@@ -1408,12 +1644,16 @@ class NotebookApp(JupyterApp):
 
     def write_server_info_file(self):
         """Write the result of server_info() to the JSON file info_file."""
-        with open(self.info_file, 'w') as f:
-            json.dump(self.server_info(), f, indent=2, sort_keys=True)
+        try:
+            with open(self.info_file, 'w') as f:
+                json.dump(self.server_info(), f, indent=2, sort_keys=True)
+        except OSError as e:
+            self.log.error(_("Failed to write server-info to %s: %s"),
+                           self.info_file, e)
 
     def remove_server_info_file(self):
         """Remove the nbserver-<pid>.json file created for this server.
-        
+
         Ignores the error raised when the file has already been removed.
         """
         try:
@@ -1441,14 +1681,14 @@ class NotebookApp(JupyterApp):
                 self.exit(1)
 
         info = self.log.info
-        for line in self.notebook_info().split("\n"):
+        for line in self.notebook_info(kernel_count=False).split("\n"):
             info(line)
         info(_("Use Control-C to stop this server and shut down all kernels (twice to skip confirmation)."))
         if 'dev' in notebook.__version__:
             info(_("Welcome to Project Jupyter! Explore the various tools available"
                  " and their corresponding documentation. If you are interested"
                  " in contributing to the platform, please visit the community"
-                 "resources section at http://jupyter.org/community.html."))
+                 "resources section at https://jupyter.org/community.html."))
 
         self.write_server_info_file()
 
@@ -1483,7 +1723,7 @@ class NotebookApp(JupyterApp):
                 '\n',
                 'Copy/paste this URL into your browser when you connect for the first time,',
                 'to login with a token:',
-                '    %s' % url_concat(self.connection_url, {'token': self.token}),
+                '    %s' % self.display_url,
             ]))
 
         self.io_loop = ioloop.IOLoop.current()
