@@ -10,9 +10,10 @@
 
 
 define([
-    'base/js/namespace',
     'jquery',
+    'base/js/namespace',
     'base/js/utils',
+    'base/js/i18n',
     'base/js/keyboard',
     'services/config',
     'notebook/js/cell',
@@ -22,9 +23,11 @@ define([
     'codemirror/lib/codemirror',
     'codemirror/mode/python/python',
     'notebook/js/codemirror-ipython'
-], function(IPython,
+], function(
     $,
+    IPython,
     utils,
+    i18n,
     keyboard,
     configmod,
     cell,
@@ -98,7 +101,7 @@ define([
         this.tooltip = options.tooltip;
         this.config = options.config;
         this.class_config = new configmod.ConfigWithDefaults(this.config,
-                                        CodeCell.config_defaults, 'CodeCell');
+                                        CodeCell.options_default, 'CodeCell');
 
         // create all attributed in constructor function
         // even if null for V8 VM optimisation
@@ -110,7 +113,7 @@ define([
         this.completer = null;
 
         Cell.apply(this,[{
-            config: $.extend({}, CodeCell.options_default), 
+            config: options.config, 
             keyboard_manager: options.keyboard_manager, 
             events: this.events}]);
 
@@ -125,11 +128,7 @@ define([
     CodeCell.options_default = {
         cm_config : {
             extraKeys: {
-                "Tab" :  "indentMore",
-                "Shift-Tab" : "indentLess",
                 "Backspace" : "delSpaceToPrevTabStop",
-                "Cmd-/" : "toggleComment",
-                "Ctrl-/" : "toggleComment"
             },
             mode: 'text',
             theme: 'ipython',
@@ -147,8 +146,6 @@ define([
         },
     };
 
-    CodeCell.config_defaults = CodeCell.options_default;
-
     CodeCell.msg_cells = {};
 
     CodeCell.prototype = Object.create(Cell.prototype);
@@ -163,6 +160,15 @@ define([
 
         var input = $('<div></div>').addClass('input');
         this.input = input;
+
+        var run_this_cell = $('<div></div>').addClass('run_this_cell');
+        run_this_cell.prop('title', 'Run this cell');
+        run_this_cell.append('<i class="fa-step-forward fa"></i>');
+        run_this_cell.click(function (event) {
+            event.stopImmediatePropagation();
+            that.execute();
+        });
+
         var prompt = $('<div/>').addClass('prompt input_prompt');
         var inner_cell = $('<div/>').addClass('inner_cell');
         this.celltoolbar = new celltoolbar.CellToolbar({
@@ -183,16 +189,18 @@ define([
         this.code_mirror.on('keydown', $.proxy(this.handle_keyevent,this));
         $(this.code_mirror.getInputField()).attr("spellcheck", "false");
         inner_cell.append(input_area);
-        input.append(prompt).append(inner_cell);
+        input.append(run_this_cell).append(prompt).append(inner_cell);
 
         var output = $('<div></div>');
         cell.append(input).append(output);
         this.element = cell;
         this.output_area = new outputarea.OutputArea({
-            selector: output, 
-            prompt_area: true, 
-            events: this.events, 
-            keyboard_manager: this.keyboard_manager});
+            config: this.config,
+            selector: output,
+            prompt_area: true,
+            events: this.events,
+            keyboard_manager: this.keyboard_manager,
+        });
         this.completer = new completer.Completer(this, this.events);
     };
 
@@ -204,6 +212,12 @@ define([
         this.element.focusout(
             function() { that.auto_highlight(); }
         );
+
+        this.events.on('kernel_restarting.Kernel', function() {
+            if (that.input_prompt_number === '*') {
+              that.set_input_prompt();
+            }
+        });
     };
 
 
@@ -308,15 +322,24 @@ define([
      */
     CodeCell.prototype.execute = function (stop_on_error) {
         if (!this.kernel) {
-            console.log("Can't execute cell since kernel is not set.");
+            console.log(i18n.msg._("Can't execute cell since kernel is not set."));
             return;
         }
 
         if (stop_on_error === undefined) {
-            stop_on_error = true;
+            if (this.metadata !== undefined && 
+                    this.metadata.tags !== undefined) {
+                if (this.metadata.tags.indexOf('raises-exception') !== -1) {
+                    stop_on_error = false;
+                } else {
+                    stop_on_error = true;
+                }
+            } else {
+               stop_on_error = true;
+            }
         }
 
-        this.output_area.clear_output(false, true);
+        this.clear_output(false, true);
         var old_msg_id = this.last_msg_id;
         if (old_msg_id) {
             this.kernel.clear_callbacks_for_msg(old_msg_id);
@@ -337,6 +360,14 @@ define([
         CodeCell.msg_cells[this.last_msg_id] = this;
         this.render();
         this.events.trigger('execute.CodeCell', {cell: this});
+        var that = this;
+        function handleFinished(evt, data) {
+            if (that.kernel.id === data.kernel.id && that.last_msg_id === data.msg_id) {
+                    that.events.trigger('finished_execute.CodeCell', {cell: that});
+                that.events.off('finished_iopub.Kernel', handleFinished);
+              }
+        }
+        this.events.on('finished_iopub.Kernel', handleFinished);
     };
     
     /**
@@ -346,6 +377,7 @@ define([
     CodeCell.prototype.get_callbacks = function () {
         var that = this;
         return {
+            clear_on_done: false,
             shell : {
                 reply : $.proxy(this._handle_execute_reply, this),
                 payload : {
@@ -363,7 +395,7 @@ define([
                     that.output_area.handle_clear_output.apply(that.output_area, arguments);
                 }, 
             },
-            input : $.proxy(this._handle_input_request, this)
+            input : $.proxy(this._handle_input_request, this),
         };
     };
     
@@ -461,7 +493,7 @@ define([
         } else {
             ns = encodeURIComponent(prompt_value);
         }
-        return 'In&nbsp;[' + ns + ']:';
+        return '<bdi>'+i18n.msg._('In')+'</bdi>&nbsp;[' + ns + ']:';
     };
 
     CodeCell.input_prompt_continuation = function (prompt_value, lines_number) {
@@ -482,6 +514,7 @@ define([
         }
         this.input_prompt_number = number;
         var prompt_html = CodeCell.input_prompt_function(this.input_prompt_number, nline);
+
         // This HTML call is okay because the user contents are escaped.
         this.element.find('div.input_prompt').html(prompt_html);
         this.events.trigger('set_dirty.Notebook', {value: true});
@@ -503,8 +536,9 @@ define([
     };
 
 
-    CodeCell.prototype.clear_output = function (wait) {
-        this.output_area.clear_output(wait);
+    CodeCell.prototype.clear_output = function (wait, ignore_queue) {
+        this.events.trigger('clear_output.CodeCell', {cell: this});
+        this.output_area.clear_output(wait, ignore_queue);
         this.set_input_prompt();
     };
 
@@ -540,7 +574,11 @@ define([
         var outputs = this.output_area.toJSON();
         data.outputs = outputs;
         data.metadata.trusted = this.output_area.trusted;
-        data.metadata.collapsed = this.output_area.collapsed;
+        if (this.output_area.collapsed) {
+            data.metadata.collapsed = this.output_area.collapsed;
+        } else {
+            delete data.metadata.collapsed;
+        }
         if (this.output_area.scroll_state === 'auto') {
             delete data.metadata.scrolled;
         } else {
